#include <GL\glut.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "Arbol.h"
#include "../defs.h"

//--------------------------------------------------------------------------------------------------------------------------------
// Void constructor
// Parameters --> None
//--------------------------------------------------------------------------------------------------------------------------------
/*Arbol::Arbol(void):  Vertices(NULL), Hojas(NULL)
{
	cx = cy= cz=0;
}*/
//--------------------------------------------------------------------------------------------------------------------------------
// Constructor with the name of a file cotaining the data
//--------------------------------------------------------------------------------------------------------------------------------
Arbol::Arbol (const char * arbol, const char* hojas, const char* simplifi, Geometry::CREATEVERTEXDATAFUNC vdfun, Geometry::CREATEINDEXDATAFUNC idfun): 
create_vertex_data_func(vdfun==NULL?Geometry::DefaultVertexDataCreator:vdfun),
create_index_data_func(idfun==NULL?Geometry::DefaultIndexDataCreator:idfun),
vertexdata(NULL), Hojas(NULL), MinDet(NULL)
{	
	cx = cy= cz= 0;
	ReadFromFile (arbol, hojas, simplifi);

	// esto no s si devera haber ak
	indexdata->SetNumValidIndices(0);
}

//--------------------------------------------------------------------------------------------------------------------------------
// Destructor. We must deallocate the memory allocated for pointers to vertices and edges
//--------------------------------------------------------------------------------------------------------------------------------
Arbol::~Arbol (void)
{
	if (vertexdata) delete vertexdata;
	if (indexdata) delete indexdata;
	delete Hojas;
	delete MinDet;
//	delete [] normales_vert;
}


//
//
//--------------------------------------------------------------------------------------------------------------------------------
// leer informacion desde el fichero
//--------------------------------------------------------------------------------------------------------------------------------
void Arbol::ReadFromFile ( const char * arbol, const char* hojas, const char* simpli/*, char* tronco*/) // Reads a obj-model from a file
{
	    
	//VERTICES
	if ((TotVerts=LeeVertices(arbol))==0) ThrowFileNotFoundError((const char*)arbol);
	if ((nHojas = LeoHojas (hojas))==0) ThrowFileNotFoundError((const char*)hojas);
//	TotTronco = LeoTronco (tronco);
//	CalculaNormalesVertice();	
	if ((TotHojas = LeoHojasSimpl(simpli))==0) ThrowFileNotFoundError((const char*)simpli);	
	RellenoRaiz();

	CalculaTexCoordsYNorms();
}


#include "TMatrix.h"

//--------------------------------------------------------------------------------------------------------------------------------
// Procesar lnea para extraer los VERTICES DE VERTICES.OBJ
//--------------------------------------------------------------------------------------------------------------------------------
int Arbol::LeeVertices ( const char *  arbol)

{
		FILE* fp_origen;
		char linea[256], str[20];
		float x,y,z;
		int countv=0;
		bool fin_vertices = false;

/*		TMatrix rot90mat;
		rot90mat.RotateX(-90);
*/		
		//cuento todos los vertices, y relleno la estructura
		if ((fp_origen = fopen (arbol, "r")) == NULL)
		{
			printf ("No he podido abrir el fichero %s\n", arbol);
			return 0;
		}
		else
		{
			while ((fgets (linea, 255, fp_origen) != NULL) && (fin_vertices == false))
		
				
				switch (linea[0])
				{
					case '#':
						switch ( linea[2])
						{
						case 'v':
							sscanf(linea,"# %s %lu",  str, &countv);
							if ( countv > 0){
									//Vertices  = new Point3D[2*countv];
									//tex_coords=new float[countv*2];
									vertexdata = create_vertex_data_func(2*countv);
									Hojas = new Hoja[countv*2];
									indexdata = create_index_data_func(countv*2*3); // 3 indices x 2 triangulos x hoja
									vertexdata->Begin();
									//para impostores

							}
							countv=0;
							break;
						
						default:
							break;

						}
						break;
					case 'v':
						switch (linea[1])
						{
							case ' ':
								{
							sscanf (linea, "v %f %f %f ", &x, &y, &z);
							//Vertices[countv].Update( x, y, z); para girarlo
							//Vertices[countv].Update( x, z, y);

//							Vector rotv(x,y,z);
//							rotv = rot90mat * rotv;
							
							vertexdata->SetVertexCoord(countv,x,z,-y);
							
							countv++;
								break;
								}
							case 't':
							// PASAMOS DE ESTO: LAS COORDENADAS DE TEXTURA DE LAS HOJAS SE CALCULAN A MANO Y LAS DEL TRONCO NO SE NECESITAN AKI
							sscanf (linea, "vt %f %f", &x, &y);
							//printf ("%f %f:%s", x, y,linea);
//							tex_coords[countv*2]=x;
//							tex_coords[countv*2+1]=y;
							//vertexdata->Begin();
//							vertexdata->SetVertexTexCoord(countv-1,x,y); // -1 porque al cargar el vertice antes se incrementa countv
							//vertexdata->End();

								break;

							default :
								break;
						}
						break;


					default:
						break;
	
				}

		fclose (fp_origen);
		}

		vertexdata->End();
		return ( countv);

		
}

//--------------------------------------------------------------------------------------------------------------------------------
// Calcular las normales de las hojas
//--------------------------------------------------------------------------------------------------------------------------------

void Arbol::GetNormalH (Hoja &aHoja)
{

	float onex, oney, onez;
	float twox, twoy, twoz;
	float threex, threey, threez;

/*	Vertices[aHoja.Vert_Hoja[0]].GetCoordinates (onex, oney, onez);
	Vertices[aHoja.Vert_Hoja[1]].GetCoordinates(twox, twoy, twoz);
	Vertices[aHoja.Vert_Hoja[2]].GetCoordinates (threex, threey, threez);*/

	vertexdata->GetVertexCoord(aHoja.Vert_Hoja[0],onex,oney,onez);
	vertexdata->GetVertexCoord(aHoja.Vert_Hoja[1],twox,twoy,twoz);
	vertexdata->GetVertexCoord(aHoja.Vert_Hoja[2],threex,threey,threez);

	float v1[3]={twox-onex,twoy-oney,twoz-onez};
	float v2[3]={threex-onex,threey-oney,threez-onez};

	Normalize(v1,v1);
	Normalize(v2,v2);
	
//	aHoja.Normal[0] = (twoz-onez)*(threey-oney) - (twoy-oney)*(threez-onez);
//	aHoja.Normal[1] = (twox-onex)*(threez-onez) - (threex-onex)*(twoz-onez);
//	aHoja.Normal[2] = (threex-onex)*(twoy-oney) - (twox-onex)*(threey-oney);

	CrossProduct(v1,v2,aHoja.Normal);

	
}



//--------------------------------------------------------------------------------------------------------------------------------
// Calcular las normales del tronco
//--------------------------------------------------------------------------------------------------------------------------------

/*void Arbol::GetNormalT (Tronco &aTronco)
{

	float onex, oney, onez;
	float twox, twoy, twoz;
	float threex, threey, threez;

	Vertices[aTronco.Vert_Tronco[0]].GetCoordinates (onex, oney, onez);
	Vertices[aTronco.Vert_Tronco[1]].GetCoordinates(twox, twoy, twoz);
	Vertices[aTronco.Vert_Tronco[2]].GetCoordinates (threex, threey, threez);
	
//	aTronco.Normal[0] = (twoz-onez)*(threey-oney) - (twoy-oney)*(threez-onez);
//	aTronco.Normal[1] = (twox-onex)*(threez-onez) - (threex-onex)*(twoz-onez);
//	aTronco.Normal[2] = (threex-onex)*(twoy-oney) - (twox-onex)*(threey-oney);

	
}
*/

void Arbol::CrossProduct(const float *v1, const float *v2, float *res)
{
	res[0] = v1[1]*v2[2] - v1[2]*v2[1];
	res[1] = v1[2]*v2[0] - v1[0]*v2[2];
	res[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

void Arbol::Normalize(const float *v, float *res)
{
	float module=sqrtf(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
	res[0]=v[0]/module;
	res[1]=v[1]/module;
	res[2]=v[2]/module;
}

/*void Arbol::CalculaNormalesVertice(void)
{
	int l;
	normales_vert = new float[TotVerts*3];
	unsigned int i,j;
	float onex, oney, onez;
	float twox, twoy, twoz;
	float threex, threey, threez;
	unsigned int iv[3];


	for (i=0; i<TotVerts; i++)
	{
		normales_vert[i*3+0] = 0.0f;
		normales_vert[i*3+1] = 0.0f;
		normales_vert[i*3+2] = 0.0f;
	}

	
	for (i=0; i<TotTronco; i++)
	{
		//unsigned int iv[3] = { Troncos[i].Vert_Tronco[0], Troncos[i].Vert_Tronco[1], Troncos[i].Vert_Tronco[3] };
		iv[0]=Troncos[i].Vert_Tronco[0];
		iv[1]=Troncos[i].Vert_Tronco[1];
		iv[2]=Troncos[i].Vert_Tronco[2];

		float auxnormal[3]={0.0f, 0.0f, 0.0f};

//		if (iv[0] != iv[1] && iv[0] != iv[2] && iv[1] != iv[2])
//		{

			Vertices[iv[0]].GetCoordinates (onex, oney, onez);
			Vertices[iv[1]].GetCoordinates(twox, twoy, twoz);
			Vertices[iv[2]].GetCoordinates (threex, threey, threez);

			float vd1[3] = {	twox - onex, 
								twoy - oney, 
								twoz - onez };

			float vd2[3] = {	threex - onex, 
								threey - oney, 						
								threez - onez };

			Normalize(vd1,vd1);
			Normalize(vd2,vd2);
//			if (i%2)
				CrossProduct(vd2,vd1,auxnormal);
//			else
//				CrossProduct(vd1,vd2,auxnormal);

//		}

		for (j=0; j<3; j++)
		{
			normales_vert[iv[j]*3+0] += auxnormal[0];
			normales_vert[iv[j]*3+1] += auxnormal[1];
			normales_vert[iv[j]*3+2] += auxnormal[2];
		}
	}

		
	for (i=0; i<TotVerts; i++)
	{
		if ((normales_vert[i*3]!=0.0)||(normales_vert[i*3+1]!=0.0)||(normales_vert[i*3+2]!=0.0))
			Normalize(normales_vert+i*3, normales_vert+i*3);
		
	}

	
	for (i=0; i<TotTronco; i++)
	{
		iv[0]=Troncos[i].Vert_Tronco[0];
		iv[1]=Troncos[i].Vert_Tronco[1];
		iv[2]=Troncos[i].Vert_Tronco[2];


		Troncos[i].Normal[0][0]=normales_vert[iv[0]*3];
		Troncos[i].Normal[0][1]=normales_vert[iv[0]*3+1];
		Troncos[i].Normal[0][2]=normales_vert[iv[0]*3+2];
		Troncos[i].Normal[1][0]=normales_vert[iv[1]*3];
		Troncos[i].Normal[1][1]=normales_vert[iv[1]*3+1];
		Troncos[i].Normal[1][2]=normales_vert[iv[1]*3+2];
		Troncos[i].Normal[2][0]=normales_vert[iv[2]*3];
		Troncos[i].Normal[2][1]=normales_vert[iv[2]*3+1];
		Troncos[i].Normal[2][2]=normales_vert[iv[2]*3+2];
	}

	delete [] normales_vert;

}*/

//--------------------------------------------------------------------------------------------------------------------------------
// Procesar lnea para extraer las HOJAS DE HOJAS.OBJ
//--------------------------------------------------------------------------------------------------------------------------------

  int Arbol::LeoHojas (const char* hojas)
{
	FILE* fp_hojas;
	char linea[256];

	int v1, v2, v3, h=0;
	


	if ((fp_hojas = fopen (hojas, "r")) == NULL)
	{
			printf ("No he podido abrir el fichero %s\n", hojas);
			return 0;
	}		
	else
	{


		while (fgets (linea, 255, fp_hojas) != NULL)
		
			switch (linea[0])
			{
				case 'H':
				
					fgets (linea, 255, fp_hojas);
					sscanf(linea, "f %lu %lu %lu", &v1, &v2, &v3);


					Hojas[h].Vert_Hoja[0]=(v1-1);
					Hojas[h].Vert_Hoja[1]=(v2-1);
					Hojas[h].Vert_Hoja[2]=(v3-1);
					Hojas[h].visible = 0;

					
					fgets (linea, 255, fp_hojas);
					sscanf(linea, "f %lu %lu %lu", &v1, &v2, &v3);
					Hojas[h].Vert_Hoja[3]=(v3-1);

					GetNormalH ( Hojas[h]);

					h++;
					break;
					
				default :
					printf( "Hay un error al leer las hojas\n");
					break;
	
			}
		
	}

	return (h);	
}


//--------------------------------------------------------------------------------------------------------------------------------
// Procesar lnea para extraer los triangulos del tronco de TRONCO.OBJ
//--------------------------------------------------------------------------------------------------------------------------------
/*
  int Arbol::LeoTronco (char* tronco)
{
	FILE* fp_tronco;
	char linea[256];
    int v1, v2, v3, t=0;
	


	if ((fp_tronco = fopen (tronco, "r")) == NULL)
			printf ("No he podido abrir el fichero %s\n", tronco);
		
	else
	{

		while (fgets (linea, 255, fp_tronco) != NULL)
		
		{
			sscanf(linea, "f %lu %lu %lu", &v1, &v2, &v3);

			Troncos[t].Vert_Tronco[0]=(v1-1);
			Troncos[t].Vert_Tronco[1]=(v2-1);
			Troncos[t].Vert_Tronco[2]=(v3-1);

			GetNormalT (Troncos[t]);
			t++;
		
		}
		
			fclose(fp_tronco);
	}
	return (t);
}
*/
//---------------------------------------------------------------------------------------------------------------------
// Procesar lnea para extraer las HOJAS DE SIMPLIFICA.OBJ
//---------------------------------------------------------------------------------------------------------------------

  int Arbol::LeoHojasSimpl (const char* simplifica)
{
	FILE* fp_simpli;
	char linea[256], str[10];;
	int v0, v1, v2, v3, tn, tv1,tv2, e=0;
	float xc,yc, zc;
	float xmax, xmin, ymax, ymin, zmax, zmin;


	if ((fp_simpli = fopen (simplifica, "r")) == NULL)
	{
			printf ("No he podido abrir el fichero %s\n", simplifica);
			return 0;
	}
		
	else
	{

		while (fgets (linea, 255, fp_simpli) != NULL)
		{
			switch (linea[0])
			{

			case 'C' :
				sscanf(linea, "%s %f %f %f ", &str, &xc, &yc, &zc );
				cx = xc;
				cy = zc;
				cz = yc;
			
				break;

			//para almacenar la caja envolvente tambin cambio las componentes y y z
			case 'x' :
				sscanf(linea, "%s %f %f", &str, &xmax, &xmin );
				break;

			case 'y' :
				sscanf(linea, "%s %f %f", &str, &zmax, &zmin );
				break;

			case 'z' :
				sscanf(linea, "%s %f %f", &str, &ymax, &ymin );
				rad = (float)(sqrt ((xmax-cx)*(xmax-cx)+ (ymax-cy)*(ymax-cy)+ (zmax-cz)*(zmax-cz)))-3;

				break;



			case 'N':
			
				//N 446 Ver 10176 10178 10169 10171  Tv 156 154 E 2
	
				sscanf(linea, "%s %lu %s %lu %lu  %lu %lu %s %lu %lu %s %i", 
					&str, &tn, &str, &v0, &v1, &v2, &v3, &str, &tv1, &tv2, &str, &e );
				Hojas[tn].Vert_Hoja[0] = v0;
				Hojas[tn].Vert_Hoja[1] = v1;
				Hojas[tn].Vert_Hoja[2] = v2;
				Hojas[tn].Vert_Hoja[3] = v3;

				Hojas[tn].visible = 0;

				GetNormalH  (Hojas[tn]);
	
				Hojas[tn].hijoi= tv1;
				Hojas[tn].hijod= tv2;

				Hojas[tv1].padre = tn;
				Hojas[tv2].padre = tn;

				break;

			default:
				break;

			}
		}

	}

		
	return (tn);
}

 
//--------------------------------------------------------------------------------------------------------------------------------
// relleno la lista de las hojas con el menor detalle para hacer el culling
// cada hoja ahora tiene un ndice a la hoja raz de su rbol
//--------------------------------------------------------------------------------------------------------------------------------

void Arbol::RellenoRaiz (void)
{
	int i,j, k, t, cont;
	bool esta, fin;

	i=0;
	k=-1;
	cont =-1;

	MinDet  = new Activas[nHojas*2];
	
	while (i<TotHojas)
	{
		j=i;
		while (Hojas[j].padre != -1) j= Hojas[j].padre;
		Hojas[i].raiz = j;


		// para la estructura MinDet
		if ( k == -1){
			k++;
			MinDet[k].indice = j;
			cont =k;
		}
		else
		{
			t = 0;
			esta = false;
			fin = false;

			while (( fin == false) && (esta == false))
			{	if ( MinDet[t].indice == j) esta = true;
					else t++;
				if (MinDet[t].indice == -1) fin = true;
			}

			if ( esta == false) //aado al final
			{
				cont++;
				MinDet[cont].indice = j;
			}
		}

	i++;
	}	

	minHojas = cont;

}

void Arbol::CalculaTexCoordsYNorms(void)
{	
	vertexdata->Begin();
	for (unsigned int i=0; i<nHojas; i++)
	{	
		const float* lanormal = Hojas[i].Normal;
		vertexdata->SetVertexNormal(Hojas[i].Vert_Hoja[0], lanormal[0], lanormal[1], lanormal[2]);
		vertexdata->SetVertexTexCoord(Hojas[i].Vert_Hoja[0], 0.0f, 1.0f);

		vertexdata->SetVertexNormal(Hojas[i].Vert_Hoja[1], lanormal[0], lanormal[1], lanormal[2]);
		vertexdata->SetVertexTexCoord(Hojas[i].Vert_Hoja[1], 0.0f, 0.0f);

		vertexdata->SetVertexNormal(Hojas[i].Vert_Hoja[2], lanormal[0], lanormal[1], lanormal[2]);
		vertexdata->SetVertexTexCoord(Hojas[i].Vert_Hoja[2], 1.0f, 1.0f);

		vertexdata->SetVertexNormal(Hojas[i].Vert_Hoja[3], lanormal[0], lanormal[1], lanormal[2]);
		vertexdata->SetVertexTexCoord(Hojas[i].Vert_Hoja[3], 1.0f, 0.0f);


/*		MiArbol->Vertices[aHoja.Vert_Hoja[1]].GetCoordinates (a,b,c);
		glTexCoord2f ( 0.0, 1.0);
		glVertex3f (a,b,c);
	
		MiArbol->Vertices[aHoja.Vert_Hoja[2]].GetCoordinates (a,b,c);
		glTexCoord2f ( 1.0, 0.0);
		glVertex3f (a,b,c);

		MiArbol->Vertices[aHoja.Vert_Hoja[3]].GetCoordinates (a,b,c);
		glTexCoord2f ( 1.0, 1.0);
		glVertex3f (a,b,c);*/
	}
	vertexdata->End();
}

Arbol::Arbol(Arbol *ar)
{
	nHojas=ar->nHojas;
	MinDet  = new Activas[nHojas*2];
//	for (unsigned int i=0; i<nHojas*2; i++)
//		MinDet[i]=ar->MinDet[i];
	memcpy(MinDet,ar->MinDet,sizeof(Activas)*nHojas*2);
	TotHojas=ar->TotHojas;
	cx=ar->cx;
	cy=ar->cy;
	cz=ar->cz;
	rad=ar->rad;
	minHojas=ar->minHojas;
	TotVerts=ar->TotVerts;

	create_vertex_data_func=ar->create_vertex_data_func;
	create_index_data_func=ar->create_index_data_func;
	vertexdata=create_vertex_data_func(ar->vertexdata->GetNumVertices());
	vertexdata->Begin(); 
	for (unsigned int i=0; i<vertexdata->GetNumVertices(); i++) 
	{
		float va,vb,vc;
		ar->vertexdata->GetVertexCoord(i,va,vb,vc);
		vertexdata->SetVertexCoord(i,va,vb,vc);
		ar->vertexdata->GetVertexNormal(i,va,vb,vc);
		vertexdata->SetVertexNormal(i,va,vb,vc);
	}
	vertexdata->End();
	indexdata=create_index_data_func(ar->indexdata->GetNumMaxIndices());
	indexdata->Begin(); 
	for (unsigned int i=0; i<indexdata->GetNumMaxIndices(); i++) 
		indexdata->SetIndex(i,ar->indexdata->GetIndex(i));
	indexdata->End();

	Hojas=new Hoja[vertexdata->GetNumVertices()];
//	for (unsigned int i=0; i<vertexdata->GetNumVertices(); i++) 
//		Hojas[i]=ar->Hojas[i];
	memcpy(Hojas,ar->Hojas,sizeof(Hoja)*vertexdata->GetNumVertices());

	// esto no s si devera haber ak
	indexdata->SetNumValidIndices(0);
}