#include <math.h> // Necesaria para el clculo de raices cuadradas.
#include "Point3D.h"


//---------------------------------------------------------------------------------------------------
// Constructor de copia sin parmetros
//---------------------------------------------------------------------------------------------------
Point3D::Point3D (void)
{
	coords[0] = coords[1] = coords[2] = 0.0f;
}


//---------------------------------------------------------------------------------------------------
// Constructor con tres parmetros
// Parmetros --> x, y, z son las coordenadas del punto en 3D
//---------------------------------------------------------------------------------------------------
Point3D::Point3D (float x, float y, float z)
{
	coords[0] = x;
	coords[1] = y;
	coords[2] = z;
}


//---------------------------------------------------------------------------------------------------
// Constructor de copia
// Parmetros --> p es el punto con el se hace la copia
//---------------------------------------------------------------------------------------------------
Point3D::Point3D (const Point3D & p)
{
	coords[0] = p.coords[0];
	coords[1] = p.coords[1];
	coords[2] = p.coords[2];
}


//---------------------------------------------------------------------------------------------------
// Operador de asignacin
// Parmetros --> p es el punto con los datos a copiar
// Salida --> Se devuelve una referencia al punto al que se acaban de copiar los datos
//---------------------------------------------------------------------------------------------------
Point3D & Point3D::operator = (const Point3D & p)
{
	coords[0] = p.coords[0];
	coords[1] = p.coords[1];
	coords[2] = p.coords[2];

	return *this;
}


//---------------------------------------------------------------------------------------------------
// Operador de comparacin
// Parmetros const Point3D& p : Punto con el que se est comparando
// Salida --> 'true' si los puntos son iguales
//			  'false' en caso contrario
//---------------------------------------------------------------------------------------------------
bool Point3D::operator == (const Point3D & p) const
{
	if (coords[0] == p.coords[0] &&
		coords[1] == p.coords[1] &&
		coords[2] == p.coords[2])
		return true;

	return false;
}


//---------------------------------------------------------------------------------------------------
// Operador diferencia
// Parmetros --> p es el punto con el que se compara el actual
// Salida si son distintos 'true' en caso contrario 'false'
//---------------------------------------------------------------------------------------------------
bool Point3D::operator != (const Point3D & p) const
{
	return (coords[0] != p.coords[0] ||
			coords[1] != p.coords[1] ||
			coords[2] != p.coords[2]);
}


//---------------------------------------------------------------------------------------------------
// Para actualizar las componentes del punto
// Parmetros --> const float a, const float b, const float c : las nuevas coordenadas del punto
//---------------------------------------------------------------------------------------------------
bool Point3D::Update (const float a, const float b, const float c)
{
	coords[0] = a;
	coords[1] = b;
	coords[2] = c;

	return true;
}



bool Point3D::Update (const Point3D& unPunto)
{
	*this = unPunto;

	return true;
}


//---------------------------------------------------------------------------------------------------
// Returns the point coordinates 
// Parameters --> float& a, float& b, float& c : References that recieve the coordinates
// Return --> None
//---------------------------------------------------------------------------------------------------
void Point3D::GetCoordinates (float& a, float& b, float& c) const
{
	a = coords[0];
	b = coords[1];
	c = coords[2];
}

//---------------------------------------------------------------------------------------------------
// Actualiza las coordenadas
// Parameters --> float a, float b, float c : Lo que se van a incrementar
//							las coordenadas del punto
// Return --> None
//---------------------------------------------------------------------------------------------------
void Point3D::Actualiza (float a, float b, float c) 
{
	coords[0] = coords[0] + a;
	coords[1] = coords[1] + b;
	coords[2] = coords[2] + c;
}



//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
const float * const Point3D::GetCoordinates (void) const
{
	return coords;
}


//---------------------------------------------------------------------------------------------------
// Calcula la distancia euclidea entre dos puntos
// Parmetros --> p es el punto hasta el que se calcula la distancia desde este
// Salida la distancia entre los puntos como un float
//---------------------------------------------------------------------------------------------------
float Point3D::Distance (const Point3D & p) const
{
	return ((float)sqrt ((coords[0]-p.coords[0])*(coords[0]-p.coords[0])+
						 (coords[1]-p.coords[1])*(coords[1]-p.coords[1])+
						 (coords[2]-p.coords[2])*(coords[2]-p.coords[2])));
}


//---------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------
std::ostream& operator << (std::ostream& os, Point3D & p)
{
	os << '(' << p.coords[0] << ',' << p.coords[1] << ',' << p.coords[2] << ")\n";

	return os;
}